import uuid
from decimal import Decimal

from PyQt5 import QtWidgets

from decide.qt.helpers import DoubleInput
from decide.qt.inputwindow import signals


class ActorInputModel(object):
    """
    Object containing a name and power
    """

    def __init__(self, name: str, power: Decimal):
        super().__init__()
        self.id = None
        self._name = name
        self._power = power
        self._comment = ""
        self.key = "actor_input"

        self.uuid = uuid.uuid4()

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self.set_name(value)

    @property
    def power(self):
        return self._power

    @power.setter
    def power(self, value):
        self.set_power(value)

    def set_name(self, value, silence=False):
        self._name = value

        if not silence:
            signals.actor_changed.send(self, key='name', value=value)

    def set_power(self, value, silence=False):

        self._power = value

        if not silence:
            signals.actor_changed.send(self, key='power', value=value)

    @property
    def comment(self):
        return self._comment

    @comment.setter
    def comment(self, value):
        self.set_comment(value)

    def set_comment(self, value):
        self._comment = value


class IssueInputModel(object):
    """
    Object containing a name, lower and upper bounds
    """

    def __init__(self, name: str, lower: Decimal, upper: Decimal):
        super().__init__()
        self.id = None
        self._name = name
        self._lower = lower
        self._upper = upper
        self._comment = ""
        self.key = "issue_input"

        self.uuid = uuid.uuid4()

    @property
    def name(self):
        return self._name

    @property
    def lower(self):
        return self._lower

    @property
    def upper(self):
        return self._upper

    @name.setter
    def name(self, value):
        self.set_name(value)

    @lower.setter
    def lower(self, value):
        self.set_lower(value)

    @upper.setter
    def upper(self, value):
        self.set_upper(value)

    def set_name(self, value, silence=False):
        self._name = value

        if not silence:
            signals.issue_changed.send(self, key='name', value=value)

    def set_lower(self, value, silence=False):
        self._lower = value

        if not silence:
            signals.issue_changed.send(self, key='lower', value=value)

    def set_upper(self, value, silence=False):
        self._upper = value

        if not silence:
            signals.issue_changed.send(self, key='upper', value=value)

    @property
    def comment(self):
        return self._comment

    @comment.setter
    def comment(self, value):
        self.set_comment(value)

    def set_comment(self, value):
        self._comment = value


class ActorIssueInputModel(object):

    def __init__(self, actor: ActorInputModel, issue: IssueInputModel):
        super().__init__()
        self.id = None

        self.actor = actor
        self.issue = issue

        self.actor_input = QtWidgets.QLabel(actor.name)
        self.issue_input = QtWidgets.QLabel(issue.name)

        self.power_input = QtWidgets.QLabel(str(actor.power))

        self.position_input = DoubleInput()
        self.position_input.valueChanged.connect(self.set_position)

        self.salience_input = DoubleInput()

        self.meta = dict()

        self.uuid = uuid.uuid4()

        self._power = Decimal(0.0)
        self._position = Decimal(0.0)
        self._salience = Decimal(0.0)

    @property
    def position(self):
        return self._position

    @property
    def salience(self):
        return self._salience

    @property
    def power(self):
        return self._power

    def set_position(self, value: Decimal, silence=False):
        self._position = value
        if not silence:
            signals.actor_issue_changed.send(self, key='position', value=value)

    def set_power(self, value: Decimal, silence=False):
        self._power = value
        if not silence:
            signals.actor_issue_changed.send(self, key='power', value=value)

    def set_salience(self, value: Decimal, silence=False):
        self._salience = value
        if not silence:
            signals.actor_issue_changed.send(self, key='salience', value=value)



import csv


class CsvWriter:
    """
    Helper functions to write an exchange to a .csv file
    """

    @staticmethod
    def write(filename, realized):
        """
        Write all realized exchanges to the given file
        """

        if len(realized) <= 0:
            return

        with open(filename, "w") as csvfile:
            writer = csv.writer(csvfile, delimiter=";", lineterminator="\n")

            # write heading
            writer.writerow(realized[0].csv_row(head=True))

            for exchange in realized:
                if exchange.is_valid:
                    writer.writerow(exchange.csv_row())



import logging
import os
from unittest import TestCase

from decide.model.equalgain import EqualGainModel
from decide.model.helpers import csvparser
from decide.model.helpers import database as db
from decide.model.observers.observer import Observable
from decide.model.observers.sqliteobserver import SQLiteObserver


class BaseDatabaseTestCase(TestCase):
    def setUp(self):
        logging.info("setup")
        # Bind model classes to test db. Since we have a complete list of
        # all models, we do not need to recursively bind dependencies.

        csv = csvparser.CsvParser(EqualGainModel())

        path = os.path.join(
            os.path.dirname(os.path.realpath(__file__)),
            "../../../../data/input/sample_data.txt",
        )
        self.model = csv.read(path)
        self.model.randomized_value = 0.10

        self.event_handler = Observable(model_ref=self.model, output_directory="")
        self.observer = SQLiteObserver(self.event_handler, output_directory=":memory:")


class TestHashFunctions(BaseDatabaseTestCase):
    def test_eq(self):
        """
        Shows the possibility of the hash/dictionary key implement
        """
        store = {}

        data_set = db.DataSet(name="pizza")
        store[data_set] = data_set
        self.assertEqual(store["pizza"], data_set)

        iteration = db.Iteration()
        iteration.pointer = 1

        iteration2 = db.Iteration()
        iteration2.pointer = 2

        iteration3 = db.Iteration()
        iteration3.pointer = 1

        store[iteration3] = iteration3

        self.assertNotEqual(iteration2, iteration)

        self.assertEqual(iteration3, iteration)

        self.assertEqual(store[iteration3], iteration)

        self.assertEqual(store[1], iteration)

    def test_database(self):
        created = db.DataSet.create(name="test123")

        data_set = db.DataSet.get(name="test123")

        self.assertEqual(data_set, created)

    def test_data_model(self):

        self.observer.before_repetitions(repetitions=1, iterations=1)  # TODO FIX
        self.observer.before_iterations(1)
        self.observer.before_loop(0, 1)

        # compare the ActorIssues against the database results
        actor_issue_count = sum([len(x) for x in self.model.actor_issues.values()])
        self.assertEqual(db.ActorIssue.select().count(), actor_issue_count)

        # the after_loop events captures again all the actor issues, so the amount of actorIssues is 2 times that high
        self.observer.end_loop(0, 1)
        self.assertEqual(db.ActorIssue.select().count(), actor_issue_count * 2)

    def test_exchange_database(self):
        db.connection.drop_tables([db.ExchangeActor, db.Exchange])
        db.connection.create_tables([db.ExchangeActor, db.Exchange])



# -*- coding: utf-8 -*-
import csv
from decimal import Decimal

from .. import base


class CsvParser:
    # csv row identifiers
    cA = "#A"  # A = actor

    # P = issues
    cP = "#P"

    # D = the position, salience & power of an actor of an issue
    cD = "#D"

    # M = issue dimensions
    cM = "#M"

    cI = "#/"

    # /	actor	issue	position	salience	power
    rActor = 1
    rIssue = 2
    rPosition = 3
    rSalience = 4
    rPower = 5

    def __init__(self, model_ref: "base.AbstractModel"):
        self.model_ref = model_ref

    def read(self, filename, actor_whitelist=list(), issue_whitelist=list()):
        """
        The file to read
        :param issue_whitelist:
        :param actor_whitelist:
        :param filename:
        :return:
        """

        with open(filename, "rt", encoding="utf-8", errors='replace') as csv_file:

            # guess the document format
            dialect = csv.Sniffer().sniff(csv_file.read(1024))
            csv_file.seek(0)

            reader = csv.reader(csv_file, dialect=dialect)

            for row in reader:

                if row[0] == self.cA:
                    self.parse_row_actor(row, actor_whitelist)
                elif row[0] == self.cP:
                    self.parse_row_issue(row, issue_whitelist)
                elif row[0] == self.cD:
                    self.parse_row_d(row, actor_whitelist, issue_whitelist)
                elif row[0] == self.cM:
                    self.parse_row_m(row, issue_whitelist)
                    pass

        self.create_issues()

        for issue_id, v in self.model_ref.actor_issues.items():

            issue = self.model_ref.issues[issue_id]

            for actor_name, value in self.model_ref.actor_issues[issue_id].items():
                normalized_position = issue.normalize(
                    self.model_ref.actor_issues[issue_id][actor_name].position
                )

                self.model_ref.actor_issues[issue_id][
                    actor_name
                ].position = normalized_position

        return self.model_ref

    def parse_row_actor(self, row, whitelist=list()):
        """
        Parse the actor row
        :param whitelist: list of actors to parse, or empty list to accept all
        :param row:
        """
        from decide.model.helpers.helpers import create_key

        actor_id = create_key(row[1])

        if len(whitelist) == 0 or actor_id in whitelist:
            self.model_ref.add_actor(row[1], actor_id, comment=row[3])

    def parse_row_issue(self, row, whitelist=list()):
        """
        The csv row
        :param whitelist:
        :param row:
        """
        from decide.model.helpers.helpers import create_key

        issue_id = create_key(row[1])

        if len(whitelist) == 0 or issue_id in whitelist:
            self.model_ref.add_issue(row[1], issue_id=issue_id, comment=row[3])

    def parse_row_m(self, row, whitelist=list()):
        """
        Parse the #M row containing the dimensions of the issues
        :param whitelist:
        :param row:
        """

        from decide.model.helpers.helpers import create_key

        issue_id = create_key(row[1])

        if len(whitelist) == 0 or issue_id in whitelist:

            issue = None

            for obj in self.model_ref.issues.values():
                if obj.issue_id == issue_id:
                    issue = obj

            if not isinstance(issue, base.Issue):
                raise Exception("The issue variable it not the correct type")

            value = Decimal(row[2].replace(",", "."))

            issue.expand_lower(value)
            issue.expand_upper(value)

    def create_issues(self):
        """
        Create the issues
        """
        for issue_id, actor_issues in self.model_ref.actor_issues.items():

            if issue_id in self.model_ref.issues:

                if (
                    self.model_ref.issues[issue_id].upper is None
                    or self.model_ref.issues[issue_id].lower is None
                ):

                    for actor_name, actor_issue in actor_issues.items():
                        self.model_ref.issues[issue_id].expand_lower(
                            actor_issue.position
                        )
                        self.model_ref.issues[issue_id].expand_upper(
                            actor_issue.position
                        )

                self.model_ref.issues[issue_id].calculate_delta()
                self.model_ref.issues[issue_id].calculate_step_size()

    def parse_row_d(self, row, actor_whitelist=list(), issue_whitelist=list()):
        """
        The #D row contains the data for each actor on each issue
        :param issue_whitelist:
        :param actor_whitelist:
        :param row:
        """
        from decide.model.helpers.helpers import create_key

        actor_id = create_key(row[self.rActor])
        issue_id = create_key(row[self.rIssue])

        if (len(actor_whitelist) == 0 or actor_id in actor_whitelist) and (
            len(issue_whitelist) == 0 or issue_id in issue_whitelist
        ):

            if str(row[self.rSalience]) != "0":
                self.model_ref.add_actor_issue(
                    actor=actor_id,
                    issue=issue_id,
                    power=row[self.rPower].replace(",", "."),
                    salience=row[self.rSalience].replace(",", "."),
                    position=row[self.rPosition].replace(",", "."),
                )



import copy
import logging
import os
import sys
import time
import xml.etree.cElementTree as ET
from typing import List

from PyQt5 import QtCore
from PyQt5 import QtWidgets

from decide import log_filename, decide_base_path
from decide.cli import init_model, init_output_directory, float_range
from decide.model import base
from decide.model.equalgain import EqualGainModel
from decide.model.helpers import csvparser, helpers
from decide.model.observers.exchanges_writer import ExchangesWriter
from decide.model.observers.externalities import Externalities
from decide.model.observers.issue_development import IssueDevelopment
from decide.model.observers.observer import Observable
from decide.model.observers.sqliteobserver import SQLiteObserver


class ProgramData(QtCore.QObject):
    """
    The data used for displaying
    """

    changed = QtCore.pyqtSignal()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.issues = {}
        self.actors = {}
        self.actor_issues = {}


class ProgramSettings(QtCore.QObject):
    """
    The settings for the model parameters
    """

    changed = QtCore.pyqtSignal()

    settings_file = "decide-settings.xml"

    def __init__(self, *args, **kwargs):
        super(ProgramSettings, self).__init__(*args, **kwargs)

        self.input_filename = ""
        self.output_directory = ""

        self.salience_weight = 0.4
        self.fixed_weight = 0.1

        self.randomized_value = 0.00
        self.start = 0.00
        self.step = 0.05
        self.stop = 0.50

        self.repetitions = 10
        self.iterations = 10

        self.settings_type = "xml"
        self.settings_list_separator = ";"

        self.version = 1

        self.output_sqlite = False
        self.issue_development_csv = True
        self.externalities_csv = False
        self.exchanges_csv = False
        self.voting_positions = False
        self.summary_only = True

        self.selected_actors = []
        self.selected_issues = []

    def settings_file_path(self):
        file_path = os.path.join(decide_base_path, self.settings_file)

        return file_path

    def save(self):
        if self.settings_type == "xml":
            self._save_xml()

    def _save_xml(self):

        file_path = self.settings_file_path()

        element = ET.Element("decide-settings")

        for key, value in self.__dict__.items():
            if not key.startswith("_") and hasattr(self, key):

                if isinstance(value, list):
                    value = self.settings_list_separator.join(value)

                child = ET.Element(key)
                child.text = str(value)
                element.append(child)

        ET.ElementTree(element).write(file_path)

    def load(self):
        if self.settings_type == "xml":
            self._load_xml()

    def _load_xml(self):

        file_path = os.path.join(decide_base_path, self.settings_file)

        if not os.path.exists(file_path):
            return

        for elm in ET.parse(file_path).getroot():

            if hasattr(self, elm.tag):

                attr = getattr(self, elm.tag)

                if isinstance(attr, bool):
                    setattr(self, elm.tag, elm.text == "True")
                elif isinstance(attr, int):
                    setattr(self, elm.tag, int(elm.text))
                elif isinstance(attr, float):
                    setattr(self, elm.tag, float(elm.text))
                elif isinstance(attr, list):
                    setattr(
                        self, elm.tag, str(elm.text).split(self.settings_list_separator)
                    )
                else:
                    setattr(self, elm.tag, str(elm.text))


class DynamicFormLayout(QtWidgets.QGridLayout):
    def __init__(self, main_window, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.main_window = main_window
        self.checkboxes = []
        self.objects = []
        self.row_pointer = 0

    def clear(self):
        """
        Set the parent of all child widgets to None so they will be deleted
        """
        for i in reversed(range(self.count())):
            self.itemAt(i).widget().setParent(None)

    def add_row(self, *widgets):

        for index, widget in enumerate(widgets):
            self.addWidget(widget, self.row_pointer, index)

        self.row_pointer += 1

    def set_values(self, objects):
        self.objects = objects
        self.objects.sort()
        self.clear()

    def get_selected(self):

        selected = []

        for checkbox in self.checkboxes:  # type: QtWidgets.QCheckBox
            if checkbox.isChecked():
                selected.append(checkbox.objectName())

        return selected


class IssueWidget(DynamicFormLayout):
    """
    Form layout for customizing the Issue selection
    """

    def set_issues(self, issues: List[base.Issue]):
        self.set_values(issues)

        for issue in issues:
            checkbox = QtWidgets.QCheckBox(issue.name)

            lower = QtWidgets.QLabel(str(round(issue.lower, 3)))
            upper = QtWidgets.QLabel(str(round(issue.upper, 3)))

            checkbox.setObjectName(issue.issue_id)
            checkbox.stateChanged.connect(self.state_changed)

            if (
                issue in self.main_window.settings.selected_issues
                or len(self.main_window.settings.selected_issues) == 0
            ):
                checkbox.setChecked(True)

            info = QtWidgets.QPushButton("values")
            info.setObjectName(issue.issue_id)
            info.clicked.connect(self.main_window.update_actor_issue_widget)

            self.checkboxes.append(checkbox)
            self.add_row(checkbox, lower, upper, info)

    def state_changed(self):
        self.main_window.overview_widget.update_widget()

    def save(self):
        self.main_window.settings.selected_issues = self.get_selected()


class ActorWidget(DynamicFormLayout):
    """
    Form layout for customizing the Actor selection
    """

    def set_actors(self, actors: List[base.Actor]):
        self.set_values(actors)

        for actor in actors:
            checkbox = QtWidgets.QCheckBox(str(actor))
            checkbox.setObjectName(actor.actor_id)

            if (
                actor in self.main_window.settings.selected_actors
                or len(self.main_window.settings.selected_actors) == 0
            ):
                checkbox.setChecked(True)

            checkbox.stateChanged.connect(self.state_changed)

            self.checkboxes.append(checkbox)
            self.add_row(checkbox)

    def state_changed(self):
        self.main_window.overview_widget.update_widget()

    def save(self):
        self.main_window.settings.selected_actors = self.get_selected()


class ActorIssueWidget(DynamicFormLayout):
    def set_actor_issues(self, issue: base.Issue, actor_issues=List[base.ActorIssue]):
        self.set_values(actor_issues)

        self.add_row(QtWidgets.QLabel(str(issue)))

        heading = ["Actor", "Position", "Salience", "Power"]

        self.add_row(*[QtWidgets.QLabel(x) for x in heading])

        for actor_issue in actor_issues:
            actor = QtWidgets.QLabel(str(actor_issue.actor))
            position = QtWidgets.QLabel(str(round(actor_issue.position, 2)))
            salience = QtWidgets.QLabel(str(actor_issue.salience))
            power = QtWidgets.QLabel(str(actor_issue.power))

            self.add_row(actor, position, salience, power)


class SettingsFormWidget(QtWidgets.QFormLayout):
    """
    FormLayout containing the different parameters for the model
    """

    def __init__(self, settings, main_window, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.settings = settings
        self.main_window = main_window

        self.fixed_weight = QtWidgets.QDoubleSpinBox()
        self.fixed_weight.setSingleStep(0.05)
        self.fixed_weight.setDecimals(2)
        self.fixed_weight.setValue(0.10)
        self.fixed_weight.setMaximum(1)

        self.salience_weight = QtWidgets.QDoubleSpinBox()
        self.salience_weight.setSingleStep(0.05)
        self.salience_weight.setMaximum(1)
        self.salience_weight.setDecimals(2)
        self.salience_weight.setValue(0.40)

        self.randomized_value = QtWidgets.QDoubleSpinBox()
        self.randomized_value.setSingleStep(0.05)

        self.start = QtWidgets.QDoubleSpinBox()
        self.start.setSingleStep(0.05)
        self.start.valueChanged.connect(self.state_changed)

        self.step = QtWidgets.QDoubleSpinBox()
        self.step.setSingleStep(0.05)
        self.step.valueChanged.connect(self.state_changed)

        self.stop = QtWidgets.QDoubleSpinBox()
        self.stop.setSingleStep(0.05)
        self.stop.valueChanged.connect(self.state_changed)

        self.iterations = QtWidgets.QSpinBox()
        self.iterations.setMinimum(1)
        self.iterations.setValue(10)
        self.iterations.setMaximum(10000)

        self.repetitions = QtWidgets.QSpinBox()
        self.repetitions.setMinimum(1)
        self.repetitions.setValue(10)
        self.repetitions.setMaximum(10000)

        self.addRow(QtWidgets.QLabel("Fixed weight"), self.fixed_weight)
        self.addRow(QtWidgets.QLabel("Salience weight"), self.salience_weight)
        self.addRow(QtWidgets.QLabel(""))
        self.addRow(QtWidgets.QLabel("Negotiation rounds"), self.iterations)
        self.addRow(QtWidgets.QLabel("Simulation repetitions"), self.repetitions)
        self.addRow(QtWidgets.QLabel(""))

        self.addRow(QtWidgets.QLabel("p-value"))
        self.addRow(QtWidgets.QLabel("Start"), self.start)
        self.addRow(QtWidgets.QLabel("Step"), self.step)
        self.addRow(QtWidgets.QLabel("Stop"), self.stop)


    def load(self):
        """
        Copy the values from the ProgramSettings object
        """

        settings = self.settings.__dict__.items()

        for key, value in settings:
            if hasattr(self, key):
                attr = getattr(self, key)

                if isinstance(value, bool):  # type: QtWidgets.QAction
                    attr.setChecked(value)
                if isinstance(attr, list):
                    self.settings.__dict__[attr] = value
                else:
                    attr.setValue(value)

    def save(self):
        """
        Set the attributes to the ProgramSettings object
        """
        settings = self.__dict__.items()

        for key, value in settings:
            if hasattr(self.settings, key):
                setattr(self.settings, key, value.value())

    def state_changed(self):
        self.main_window.overview_widget.update_widget()

class Worker(QtCore.QObject):
    finished = QtCore.pyqtSignal(str, int)
    update = QtCore.pyqtSignal(int, int, float)

    def __init__(self, settings: ProgramSettings):
        super(Worker, self).__init__()

        self.settings = settings
        self._break = False

    @QtCore.pyqtSlot()
    def run_model(self):
        settings = self.settings

        print("start")
        print(settings.start)

        if settings.start == settings.stop:
            p_values = [settings.start]
        else:
            p_values = [
                str(round(p, 2))
                for p in float_range(
                    start=settings.start, step=settings.step, stop=settings.stop
                )
            ]

        print("P values")
        print(p_values)

        selected_actors = settings.selected_actors
        selected_issues = settings.selected_issues

        repetitions = settings.repetitions
        iterations = settings.iterations

        input_filename = settings.input_filename

        for p in p_values:
            model = init_model("equal", settings.input_filename, p=p)

            csv_parser = csvparser.CsvParser(model)
            csv_parser.read(
                input_filename,
                actor_whitelist=selected_actors,
                issue_whitelist=selected_issues,
            )

            output_directory = init_output_directory(
                model=model,
                output_dir=settings.output_directory,
                selected_actors=settings.selected_actors,
            )

            actor_issues = copy.deepcopy(model.actor_issues)

            event_handler = init_event_handlers(model, output_directory, settings)
            event_handler.before_repetitions(
                repetitions=repetitions, iterations=iterations
            )

            start_time = time.time()

            for repetition in range(repetitions):

                model.actor_issues = copy.deepcopy(actor_issues)

                model_loop = helpers.ModelLoop(model, event_handler, repetition)

                event_handler.before_iterations(repetition)

                for iteration_number in range(iterations):

                    if self._break:
                        break

                    logging.info("round {0}.{1}".format(repetition, iteration_number))
                    self.update.emit(repetition, iteration_number, start_time)

                    model_loop.loop()

                event_handler.after_iterations(repetition)

                if self._break:
                    break

            event_handler.after_repetitions()

            self.finished.emit(output_directory, model.tie_count)

    def stop(self):
        self._break = True


class SummaryWidget(DynamicFormLayout):
    def __init__(
        self, main_window, settings, data, actor_widget, issue_widget, *args, **kwargs
    ):
        """

        :type settings: ProgramSettings
        :type data: ProgramData
        :type actor_widget: ActorWidget
        :type issue_widget: IssueWidget
        """
        super(SummaryWidget, self).__init__(main_window, *args, **kwargs)

        self.settings = settings
        self.data = data
        self.actor_widget = actor_widget
        self.issue_widget = issue_widget

    def update_widget(self):

        self.clear()

        self.add_text_row("Input", self.settings.input_filename, self.test_callback)
        self.add_text_row(
            "Output directory", self.settings.output_directory, self.test_callback
        )

        settings = self.main_window.settings  # type: ProgramSettings
        settings_widget = self.main_window.settings_widget  # type: SettingsFormWidget

        p_values = [
            str(round(p, 2))
            for p in float_range(
                start=float(settings_widget.start.value()),
                step=float(settings_widget.step.value()),
                stop=float(settings_widget.stop.value()),
            )
        ]

        self.add_text_row("p-values", ", ".join(p_values))

        output_selection = []

        if settings.issue_development_csv:
            output_selection.append("Issue development values [.csv]")

        if settings.summary_only:
            output_selection.append("Summary only [.csv]")

        if settings.output_sqlite:
            output_selection.append("Sqlite [database]")

        if settings.exchanges_csv:
            output_selection.append("Exchange values [.csv]")

        if settings.externalities_csv:
            output_selection.append("Externalities [.csv]")

        if settings.voting_positions:
            output_selection.append("Show Voting positions [.csv]")

        self.add_text_row("Output settings", output_selection[0])

        for setting in output_selection[-1:]:
            self.add_text_row("", setting)

        self.main_window.set_start_button_state()

    def add_text_row(self, label, value, callback=None):

        if callback:
            value_label = QtWidgets.QLabel(
                '<a href="{}">...{}/</a>'.format(value, value[-50:])
            )
            value_label.linkActivated.connect(callback)
        else:
            value_label = QtWidgets.QLabel(value)

        self.add_row(QtWidgets.QLabel(label), value_label)

    def test_callback(self, link):
        helpers.open_file(link)


class MenuBar(QtWidgets.QMenuBar):
    def __init__(self, main_window, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.main_window = main_window
        self.settings = main_window.settings

        file_menu = self.addMenu("&File")
        output_menu = self.addMenu("&Output")

        self.output_sqlite = QtWidgets.QAction("&sqlite", self)
        self.output_sqlite.setCheckable(True)

        self.issue_development_csv = QtWidgets.QAction("Issue development (.csv)", self)
        self.issue_development_csv.setCheckable(True)
        self.issue_development_csv.setChecked(True)

        self.externalities_csv = QtWidgets.QAction("Externalities (.csv)", self)
        self.externalities_csv.setCheckable(True)

        self.exchanges_csv = QtWidgets.QAction("Exchanges (.csv)", self)
        self.exchanges_csv.setCheckable(True)

        self.voting_positions = QtWidgets.QAction("Show voting positions", self)
        self.voting_positions.setCheckable(True)

        self.summary_only = QtWidgets.QAction("Summary only (.csv)", self)
        self.summary_only.setCheckable(True)
        self.summary_only.setChecked(True)

        self.open_data_view = QtWidgets.QAction("Data view", self)
        self.open_data_view.triggered.connect(self.main_window.open_data_view)

        open_action = QtWidgets.QAction("&Open", self)
        open_action.triggered.connect(self.main_window.open_input_data)

        save_settings = QtWidgets.QAction("&Save settings", self)
        save_settings.triggered.connect(self.main_window.save_settings)

        file_menu.addAction(open_action)
        file_menu.addAction(save_settings)
        file_menu.addSeparator()
        file_menu.addAction(self.open_data_view)

        output_dir_action = QtWidgets.QAction("&Directory", self)
        output_dir_action.triggered.connect(self.main_window.select_output_dir)

        output_menu.addAction(self.issue_development_csv)
        output_menu.addAction(self.externalities_csv)
        output_menu.addAction(self.exchanges_csv)
        output_menu.addSeparator()
        output_menu.addAction(self.summary_only)
        output_menu.addAction(self.output_sqlite)
        output_menu.addAction(self.voting_positions)
        output_menu.addSeparator()
        output_menu.addAction(output_dir_action)

        debug = self.addMenu("Debug")
        log_action = QtWidgets.QAction("Show log window", self)
        log_action.triggered.connect(self.main_window.show_debug_dialog)
        debug.addAction(log_action)

        error_report = QtWidgets.QAction("Send error report", self)
        error_report.triggered.connect(self.main_window.show_error_report_dialog)
        debug.addAction(error_report)

        error_report_2 = QtWidgets.QAction("Trigger error", self)
        error_report_2.triggered.connect(self.trigger_error)
        debug.addAction(error_report_2)

    def trigger_error(self):
        raise Exception('Error triggered')

    def load(self):
        """
        Copy the values from the ProgramSettings object
        """

        settings = self.settings.__dict__.items()

        for key, value in settings:
            if hasattr(self, key):
                attr = getattr(self, key)

                if isinstance(value, bool):  # type: QtWidgets.QAction
                    attr.setChecked(value)
                else:
                    attr.setValue(value)

    def save(self):
        """
        Set the attributes to the ProgramSettings object
        """
        settings = self.__dict__.items()

        for key, value in settings:
            if hasattr(self.settings, key):
                attr = getattr(self, key)

                if isinstance(attr, QtWidgets.QAction):
                    setattr(self.settings, key, value.isChecked())
                else:
                    setattr(self.settings, key, value.value())


def init_event_handlers(model, output_directory, settings):
    """
    :type model: decide.model.base.AbstractModel
    :type output_directory: str
    :type settings: ProgramSettings
    """

    event_handler = Observable(model_ref=model, output_directory=output_directory)

    if settings.output_sqlite:
        SQLiteObserver(event_handler, settings.output_directory)

    if settings.externalities_csv:
        Externalities(event_handler, settings)

    if settings.exchanges_csv:
        ExchangesWriter(event_handler, settings.summary_only)

    if settings.issue_development_csv:
        IssueDevelopment(event_handler, summary_only=settings.summary_only)

    return event_handler


class DecideMainWindow(QtWidgets.QMainWindow):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.data = ProgramData()
        self.settings = ProgramSettings()
        self.menu_bar = None
        self.start = QtWidgets.QPushButton("Start")

        self.issue_widget = IssueWidget(self)
        self.actor_widget = ActorWidget(self)
        self.actor_issue_widget = ActorIssueWidget(self)
        self.settings_widget = SettingsFormWidget(self.settings, self)

        self.progress_dialog = None

        self.thread = QtCore.QThread()
        self.worker = None

        self.overview_widget = SummaryWidget(
            self, self.settings, self.data, self.actor_widget, self.issue_widget
        )

        self.init_ui()

        self.load_settings()

        self.init_ui_data()

    def show_debug_dialog(self):
        helpers.open_file(log_filename)

    def show_error_report_dialog(self):
        from decide.qt.errordialog import ErrorDialog
        ex = ErrorDialog('Error message', self)

    def update_data_widgets(self):

        self.issue_widget.set_issues(list(self.data.issues.values()))
        self.actor_widget.set_actors(list(self.data.actors.values()))

        if len(self.issue_widget.objects) > 0:
            issue = self.issue_widget.objects[0]
            actor_issues = list(self.data.actor_issues[issue].values())
            self.actor_issue_widget.set_actor_issues(issue.name, actor_issues)

        self.overview_widget.update_widget()

    def update_actor_issue_widget(self):
        """
        Button event to update the ActorIssueWidget
        """
        issue = self.data.issues[self.sender().objectName()]
        actor_issues = list(self.data.actor_issues[issue].values())

        self.actor_issue_widget.set_actor_issues(issue, actor_issues)
        self.overview_widget.update_widget()

    def open_input_data(self):
        """
        Open the file, parse the data and update the layout
        """
        file_name, _ = QtWidgets.QFileDialog.getOpenFileName(self, "Select input data")

        if file_name:
            self.statusBar().showMessage("Input file set to {} ".format(file_name))

            self.settings.input_filename = file_name

            self.init_ui_data()

        self.overview_widget.update_widget()

    def select_output_dir(self):
        """
        Output directory
        """
        self.settings.output_directory = str(
            QtWidgets.QFileDialog.getExistingDirectory(self, "Select Directory")
        )
        self.statusBar().showMessage(
            "Output directory set to: {} ".format(self.settings.output_directory)
        )

        self.overview_widget.update_widget()

    def set_start_button_state(self):
        if (
            self.settings.input_filename
            and self.settings.output_directory
            and not self.thread.isRunning()
        ):
            self.start.setEnabled(True)
            self.start.setDisabled(False)
        else:
            self.start.setEnabled(False)
            self.start.setDisabled(True)

    def init_ui(self):
        self.statusBar().showMessage("Ready")

        self.menu_bar = MenuBar(self, self)

        self.setMenuBar(self.menu_bar)

        self.menu_bar.setNativeMenuBar(False)

        main = QtWidgets.QHBoxLayout()

        self.start.clicked.connect(self.run)
        self.set_start_button_state()

        left = QtWidgets.QVBoxLayout()

        left_top = QtWidgets.QHBoxLayout()

        # left_top.setAlignment(QtCore.Qt.AlignTop)

        actor_box = QtWidgets.QGroupBox("Actors")
        issue_box = QtWidgets.QGroupBox("Issues")
        actor_issue_box = QtWidgets.QGroupBox("Actor Issues")

        actor_box.setLayout(self.actor_widget)
        issue_box.setLayout(self.issue_widget)
        actor_issue_box.setLayout(self.actor_issue_widget)

        left_top.addWidget(actor_box, 1)
        left_top.addWidget(issue_box, 1)

        left.addLayout(left_top, 1)
        left.addWidget(actor_issue_box, 1)

        settings_box = QtWidgets.QGroupBox("Model parameters")
        settings_box.setLayout(self.settings_widget)

        overview_box = QtWidgets.QGroupBox("Overview")
        overview_box.setLayout(self.overview_widget)
        self.overview_widget.setAlignment(QtCore.Qt.AlignTop)

        right = QtWidgets.QVBoxLayout()
        right.addWidget(settings_box, 1)
        right.addWidget(overview_box, 1)
        right.addWidget(self.start, 1)

        main.addLayout(left, 1)
        main.addLayout(right, 1)

        q = QtWidgets.QWidget()
        q.setLayout(main)

        self.setCentralWidget(q)

        self.setGeometry(300, 300, 1024, 768)
        self.setWindowTitle("Decide Exchange Model")
        self.show()

    def open_data_view(self):

        from decide.qt.inputwindow.gui import InputWindow, register_app
        ex = InputWindow(self)

        register_app(ex)

    def init_ui_data(self):

        if not os.path.isfile(self.settings.input_filename):
            return

        try:

            model = EqualGainModel()
            csv_parser = csvparser.CsvParser(model)
            csv_parser.read(self.settings.input_filename)

            self.data.actors = model.actors
            self.data.issues = model.issues
            self.data.actor_issues = model.actor_issues
            self.update_data_widgets()

        except Exception as e:
            self.statusBar().showMessage(str(e))
            QtWidgets.QMessageBox.about(self, "Input data invalid", str(e))
            logging.error(e)

    def run_safe(self):

        self.worker = Worker(self.settings)
        self.worker.finished.connect(self.finished)
        self.worker.moveToThread(self.thread)
        self.worker.finished.connect(self.thread.quit)
        self.worker.update.connect(self.update_progress)
        self.thread.started.connect(self.worker.run_model)
        self.thread.start()

        print("thread started")

    def update_progress(self, repetition, iteration, start_time):

        value = repetition * (self.settings.iterations) + iteration

        self.progress_dialog.setValue(value)
        repetitions = self.settings.repetitions

        if repetition > 0 and repetitions > 100 and repetition % 20 == 0:
            time_expired = time.time() - start_time

            avg_time_per_repetition = time_expired / repetition

            estimated_time_needed_seconds = avg_time_per_repetition * repetitions

            estimated_time_left = estimated_time_needed_seconds - (
                repetition * avg_time_per_repetition
            )

            self.statusBar().showMessage(
                "{:.0f} minutes remaining ".format(estimated_time_left / 60)
            )

    def finished(self, output_directory, tie_count):

        self._clean_progress_dialog()

        if not self.worker._break:
            buttonReply = QtWidgets.QMessageBox.question(
                self,
                "Done",
                "Done executing. Found {} ties. Open the result directory?".format(
                    tie_count
                ),
                QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No,
                QtWidgets.QMessageBox.No,
            )

            if buttonReply == QtWidgets.QMessageBox.Yes:
                if self.settings.summary_only and self.settings.issue_development_csv:
                    helpers.open_file(os.path.join(output_directory, "issues", "summary"))
                else:
                    helpers.open_file(output_directory)

    def _clean_progress_dialog(self):
        self.progress_dialog.setValue(
            self.settings.iterations * self.settings.repetitions
        )
        self.progress_dialog.hide()

    def run(self):
        # store the current state of the app
        self.save_settings()

        # create a whitelist for the selected actors

        selected_actors = self.actor_widget.get_selected()
        selected_actors.sort()

        actors_subset = "-".join(selected_actors)

        self.setWindowTitle("Decide Exchange Model {}".format(actors_subset))

        self.show_progress_dialog(actors_subset)

        self.run_safe()

    def show_progress_dialog(self, title):

        self.progress_dialog = QtWidgets.QProgressDialog(
            "Task in progress",
            "Cancel",
            0,
            self.settings.repetitions * self.settings.iterations,
            self,
        )

        bar = QtWidgets.QProgressBar(self.progress_dialog)
        bar.setTextVisible(True)
        bar.setFormat("%v/%m (%p%)")
        bar.setMinimum(0)
        bar.setMaximum(self.settings.iterations * self.settings.repetitions)
        self.progress_dialog.setBar(bar)
        # self.progress_dialog.setWindowTitle(title)

        self.progress_dialog.canceled.connect(self.cancel)

        self.progress_dialog.show()

    def load_settings(self):

        if os.path.isfile(self.settings.settings_file):
            try:
                self.settings.load()
                self.menu_bar.load()
                self.settings_widget.load()

            except ET.ParseError:
                logging.info("Invalid xml")

    def save_settings(self):
        self.settings_widget.save()
        self.issue_widget.save()
        self.actor_widget.save()
        self.menu_bar.save()
        self.settings.save()

    def cancel(self):
        self.worker.stop()


def main():

    logging.basicConfig(
        filename=log_filename,
        filemode="w",
        level=logging.DEBUG,
        format=" %(asctime)s - %(levelname)s - %(message)s",
    )

    sys.excepthook = helpers.exception_hook
    app = QtWidgets.QApplication(sys.argv)
    app.setQuitOnLastWindowClosed(True)

    decide = DecideMainWindow()

    sys.exit(app.exec_())


if __name__ == "__main__":
    main()



from blinker import Namespace

namespace = Namespace()

actor_created = namespace.signal('actor_created')
actor_changed = namespace.signal('actor_changed')
actor_deleted = namespace.signal('actor_deleted')

issue_created = namespace.signal('issue_created')
issue_changed = namespace.signal('issue_changed')
issue_deleted = namespace.signal('issue_deleted')

actor_issue_created = namespace.signal('actor_issue_created')
actor_issue_changed = namespace.signal('actor_issue_changed')
actor_issue_deleted = namespace.signal('actor_issue_deleted')




import os
from decimal import *
from unittest import TestCase
from unittest.mock import patch, MagicMock

from decide.model import calculations, base
from decide.model.equalgain import EqualGainExchange, EqualGainModel
from decide.model.helpers import csvparser


class BaseModelTest(TestCase):
    def setUp(self):
        self.model = EqualGainModel()

        filename = os.path.join(
            os.path.dirname(os.path.realpath(__file__)),
            "../../../data/input/sample_data.txt",
        )

        csv = csvparser.CsvParser(self.model)
        csv.read(filename=filename)


class TestNBSCalculations(TestCase):
    def setUp(self):
        a = base.Actor("a")
        b = base.Actor("b")
        c = base.Actor("c")

        p = base.Issue("p")

        self.a = base.ActorIssue(a, p, position=0, salience=1, power=1)
        self.b = base.ActorIssue(b, p, position=100, salience=1, power=1)
        self.c = base.ActorIssue(c, p, position=100, salience=1, power=1)

        self.actor_issues = {"a": self.a, "b": self.b, "c": self.c}
        self.denominator = calculations.calc_nbs_denominator(self.actor_issues)
        self.model = EqualGainModel()

    def test_calc_nbs(self):
        denominator = calculations.calc_nbs_denominator({"a": self.a, "b": self.b})

        self.assertEqual(
            calculations.nash_bargaining_solution(
                {"a": self.a, "b": self.b}, denominator
            ),
            50,
        )
        self.assertEqual(
            calculations.nash_bargaining_solution(self.actor_issues, self.denominator),
            Decimal(200) / Decimal(3),
        )

    def test_calc_adjusted_nbs(self):
        a1 = "a"
        a2 = "b"
        a3 = "c"

        self.assertEqual(
            calculations.adjusted_nbs(self.actor_issues, {}, a1, 100, self.denominator),
            100,
        )

        self.assertEqual(
            calculations.adjusted_nbs(self.actor_issues, {}, a2, 0, self.denominator),
            Decimal(100) / Decimal(3),
        )

        self.assertEqual(
            calculations.adjusted_nbs(self.actor_issues, {}, a1, 0, self.denominator),
            Decimal(200) / Decimal(3),
        )

        self.assertEqual(
            calculations.adjusted_nbs(
                self.actor_issues, {"a": 100}, a3, 0, self.denominator
            ),
            Decimal(200) / Decimal(3),
        )

    def test_by_absolute_move(self):
        csv = csvparser.CsvParser(self.model)

        file = os.path.join(
            os.path.dirname(os.path.realpath(__file__)),
            "../../../data/input/sample_data.txt",
        )
        model = csv.read(file)

        i = model.actors["wcentr"]
        j = model.actors["willems"]
        p = model.issues["tolheffingbinnenstad"]
        q = model.issues["extrainvestering"]
        e = EqualGainExchange(i, j, p, q, model, groups=["a", "d"])

        dp = calculations.by_absolute_move(model.actor_issues[e.i.supply.issue], e.i)
        dq = calculations.by_exchange_ratio(e.i, dp)

        self.assertAlmostEqual(dp, Decimal(13.15789473684210526315789474), delta=0.001)
        self.assertAlmostEqual(dq, Decimal(10.38781163434903047091412743), delta=0.001)

        move = calculations.reverse_move(
            model.actor_issues[e.j.supply.issue], exchange_ratio=dq, actor=e.j
        )

        self.assertAlmostEqual(
            move, Decimal(61.05117364047237206589881911), delta=0.001
        )

        self.assertLess(move, abs(e.i.supply.position - e.j.demand.position))

        dp_1 = calculations.by_absolute_move(model.actor_issues[e.j.supply.issue], e.j)
        dq_1 = calculations.by_exchange_ratio(e.j, dp)

        move_1 = calculations.reverse_move(
            model.actor_issues[e.i.supply.issue], e.i, dq_1
        )
        self.assertGreater(move_1, abs(e.i.supply.position - e.j.demand.position))

    def test_sumSaliencePower(self):
        self.assertEqual(
            calculations.sum_salience_power({"a": self.a, "b": self.b, "c": self.c}),
            (3),
        )

    def test_new_start_position(self):
        self.assertAlmostEqual(
            calculations.new_start_position(1, 0, 100), 10, delta=1e-10
        )
        self.assertAlmostEqual(
            calculations.new_start_position(1, 100, 0), 90, delta=1e-10
        )


class TestCalculations(BaseModelTest):
    @patch("decide.model.calculations.sum_salience_power")
    def test_reverse_move(self, sum_salience_power):
        sum_salience_power.return_value = 3
        actor = MagicMock()
        actor.supply = MagicMock(salience=1, power=1)

        res = calculations.reverse_move(actor_issues=[], actor=actor, exchange_ratio=1)

        self.assertEqual(res, 3)

    def test_sum_salience_power(self):
        actor_issues = [
            base.ActorIssue(None, None, 1, 1, 1),
            base.ActorIssue(None, None, 1, 1, 1),
        ]

        mock = MagicMock()
        mock.values = MagicMock(return_value=actor_issues)

        self.assertEqual(mock.values(), actor_issues)

        self.assertEqual(calculations.sum_salience_power(mock), 2)

        actor_issues.append(base.ActorIssue(None, None, 1, 1, 1))

        self.assertEqual(calculations.sum_salience_power(mock), 3)



import os

from decide.model.helpers.helpers import example_data_file_path


def test_example_data_file_path():

    assert example_data_file_path("kopenhagen").endswith(
        os.path.join("..", "..", "..", "data", "input", "kopenhagen.csv")
    )



import copy
import decimal
import logging
import os
from datetime import datetime

from PyQt5.QtCore import QDir

from decide.model import randomrate, equalgain
from decide.model.helpers import helpers, csvparser
from decide.model.observers.exchanges_writer import ExchangesWriter
from decide.model.observers.externalities import Externalities
from decide.model.observers.issue_development import IssueDevelopment
from decide.model.observers.logger import Logger
from decide.model.observers.observer import Observable
from decide.model.observers.sqliteobserver import SQLiteObserver


def init_model(model_type, input_file, p=None):
    """
    Initial the right model from the given arguments
    """

    if model_type == "equal":
        p = decimal.Decimal(p) if p else 0.0
        model = equalgain.EqualGainModel(randomized_value=p)
    else:
        model = randomrate.RandomRateModel()

    model.data_set_name = input_file.split("/")[-1].split(".")[0]

    return model


def init_event_handlers(model, output_directory, database_file, write_csv=True):
    event_handler = Observable(model_ref=model, output_directory=output_directory)

    SQLiteObserver(event_handler, database_file)

    Logger(event_handler)

    if write_csv:
        # csv handlers
        Externalities(event_handler)
        ExchangesWriter(event_handler)
        IssueDevelopment(event_handler)

    return event_handler


def init_output_directory(model, output_dir, selected_actors=list()):

    if len(model.actors) == len(selected_actors):
        actor_unique = "all"
    else:
        actor_unique = "-".join(selected_actors)

    output_directory = os.path.join(output_dir, model.data_set_name, model.model_name)

    if not os.path.isdir(output_directory):
        os.makedirs(output_directory)

    return output_directory


def float_range(start=0.0, stop=1.0, step=0.05):

    # when the step size is 0, return
    if step == 0:
        yield 0
    else:
        # prevent locking yourself out
        if (stop - start) / step > 256:
            raise RuntimeError("Maximum steps exceeded with step={} (256)".format(step))

        i = start
        # add a 10th step to stop for floating point rounding differences 0.500001 vs 0.499999
        while i < (stop + step / 10):
            yield i
            i += step


def main():
    args = helpers.parse_arguments()

    p_values = []

    if args.start and args.step and args.stop:
        p_values = [
            str(round(p, 2))
            for p in float_range(
                start=float(args.start), step=float(args.step), stop=float(args.stop)
            )
        ]

    if args.p:

        if ";" in args.p:
            p_values += args.p.split(";")
        else:
            p_values.append(args.p)

    issues = None

    if args.issues:
        issues = args.issues.split(
            ";"
        )  # 'commitments;control;devlopc2020;domestred;extra'.split(';')

    actors = None
    if args.actors:
        actors = args.actors.split(
            ";"
        )  # 'australia;brazil;canada;euinclnorway;japan;russia;usa'.split(';')

    print(p_values)
    print(issues)
    print(actors)

    for p in p_values:

        print(p)

        start_time = datetime.now()  # for timing operations

        model = init_model(args.model, args.input_file, p)

        output_directory = QDir.toNativeSeparators(init_output_directory(model, args.output_dir))

        # The event handlers for logging and writing the results to the disk.
        event_handler = init_event_handlers(
            model=model,
            output_directory=output_directory,
            database_file=args.database,
            write_csv=False,
        )

        event_handler.log(message="Start calculation at {0}".format(start_time))

        csv_parser = csvparser.CsvParser(model)
        csv_parser.read(args.input_file, actor_whitelist=actors, issue_whitelist=issues)

        actor_issues = copy.deepcopy(model.actor_issues)

        event_handler.log(message="Parsed file".format(args.input_file))

        event_handler.before_repetitions(
            repetitions=args.repetitions, iterations=args.iterations
        )

        for repetition in range(args.repetitions):

            model.actor_issues = copy.deepcopy(actor_issues)

            model_loop = helpers.ModelLoop(model, event_handler, repetition)

            event_handler.before_iterations(repetition)

            for iteration_number in range(args.iterations):
                logging.info("round {0}.{1}".format(repetition, iteration_number))
                model_loop.loop()

            event_handler.after_iterations(repetition)

            print("{}/{}".format(repetition, args.repetitions))

        event_handler.after_repetitions()

        event_handler.log(message="Finished in {0}".format(datetime.now() - start_time))

    print("done")


if __name__ == "__main__":
    main()



import argparse
import logging
import os
import re
import subprocess
import sys
import traceback

from PyQt5 import QtWidgets

from decide.model import base
from . import csvparser


def parse_arguments():
    parser = argparse.ArgumentParser(
        description="This program accepts input with a dot (.) as decimal separator. \n"
        "Parameters:\n{0} is for defining an actor,\n"
        "{1} for an issue,\n"
        "{2} for actor values for each issue.\n"
        "We expect for {2} the following order in values: "
        "actor, issue, position, salience, power".format(
            csvparser.CsvParser.cA, csvparser.CsvParser.cP, csvparser.CsvParser.cD
        )
    )
    parser.add_argument(
        "--model",
        "-m",
        help='The type of the model. The options are "equal" for the Equal Gain model and '
        '"random" for the RandomRate model ',
        default="equal",
        type=str,
    )

    parser.add_argument(
        "--p", "-p", help="Randomized Equal Gain", default=None, type=str
    )
    parser.add_argument(
        "--iterations",
        "-i",
        help="The number of round the model needs to be executed",
        default=10,
        type=int,
    )
    parser.add_argument(
        "--repetitions",
        "-r",
        help="How many times it has te be repeated?",
        default=1,
        type=int,
    )
    parser.add_argument(
        "--input_file",
        help="The location of the csv input file. ",
        default="../data/input/sample_data.txt",
        type=str,
    )
    parser.add_argument(
        "--output_dir", help="Output directory ", default="../data/output/", type=str
    )
    parser.add_argument(
        "--database",
        help="The SQLite database",
        default="../data/output/decide-data_1.db",
        type=str,
    )

    parser.add_argument("--step", default=None, type=str)
    parser.add_argument("--stop", default=None, type=str)
    parser.add_argument("--start", default=None, type=str)
    parser.add_argument("--actors", default=None, type=str)
    parser.add_argument("--issues", default=None, type=str)

    return parser.parse_args()


def create_key(value):
    """
    Create a safe for for index usage
    :type value: str
    """
    pattern = re.compile("[\W_]+")
    return pattern.sub("", value).lower()


class ModelLoop(object):
    def __init__(self, model, event_handler: "observer.Observable", repetition: int):
        self.model = model
        self.event_handler = event_handler
        self.iteration_number = 0
        self.repetition_number = repetition

    def loop(self):
        self.model.calc_nbs()
        self.model.determine_positions()
        self.model.calc_combinations()
        self.model.determine_groups()

        realized = []

        # call the event for beginning the loop
        self.event_handler.before_loop(self.iteration_number, self.repetition_number)

        while len(self.model.exchanges) > 0:

            realize_exchange = self.model.highest_gain()  # type: base.AbstractExchange

            if realize_exchange and realize_exchange.is_valid:
                removed_exchanges = self.model.remove_invalid_exchanges(
                    realize_exchange
                )

                realized.append(realize_exchange)

                self.event_handler.removed_exchanges(removed_exchanges)
                self.event_handler.execute_exchange(realize_exchange)
            else:
                if self.model.VERBOSE:
                    logging.info(realize_exchange)

        # call the event for ending the loop
        self.event_handler.after_loop(
            realized=realized,
            iteration=self.iteration_number,
            repetition=self.repetition_number,
        )

        for exchange in realized:
            self.model.actor_issues[exchange.i.supply.issue][
                exchange.i.actor
            ].position = exchange.i.y
            self.model.actor_issues[exchange.j.supply.issue][
                exchange.j.actor
            ].position = exchange.j.y

        # calc the new NBS on the voting positions and fire the event for ending this loop
        self.model.calc_nbs()
        self.event_handler.end_loop(
            iteration=self.iteration_number, repetition=self.repetition_number
        )

        # calculate for each realized exchange there new start positions
        for exchange in realized:
            pi = exchange.i.new_start_position()
            self.model.actor_issues[exchange.i.supply.issue][
                exchange.i.actor
            ].position = pi

            pj = exchange.j.new_start_position()
            self.model.actor_issues[exchange.j.supply.issue][
                exchange.j.actor
            ].position = pj

        self.iteration_number += 1


def example_data_file_path(filename):
    """
    :param filename:
    :type filename: str
    :return: full path to filename.csv
    :rtype: str
    """
    return os.path.join(
        os.path.dirname(os.path.realpath(__file__)),
        "..",
        "..",
        "..",
        "data",
        "input",
        "{}.csv".format(filename),
    )


def log_settings():
    """
    Reads the settings file into a string and logs it as info
    """

    settings_file = open("decide-settings.xml")

    settings_content = settings_file.read()

    logging.info(settings_content)


def open_file(path):

    if sys.platform.startswith("darwin"):
        subprocess.call(("open", path))
    elif os.name == "nt":
        os.startfile(path)
    elif os.name == "posix":
        subprocess.call(("xdg-open", path))


def exception_hook(exctype, ex, _traceback):
    """
    Setting the system exception hook so the exception will be logged and the log file displayed
    """

    tb_text = exception_to_string(ex)
    logging.exception(tb_text)

    print(tb_text)

    app = QtWidgets.QApplication(sys.argv)
    app.setQuitOnLastWindowClosed(True)

    from decide.qt.errordialog import ErrorDialog
    error_dialog = ErrorDialog(tb_text)
    sys.exit(error_dialog.exec_())


def exception_to_string(ex):
    tb_lines = traceback.format_exception(ex.__class__, ex, ex.__traceback__)
    tb_text = ''.join(tb_lines)
    return tb_text



import math
import os
from decimal import *
from unittest import TestCase

from decide.model.equalgain import EqualGainModel
from decide.model.helpers import csvparser


class TestModel(TestCase):
    def setUp(self):
        csv = csvparser.CsvParser(EqualGainModel())
        file = os.path.join(
            os.path.dirname(os.path.realpath(__file__)),
            "../../../data/input/sample_data.txt",
        )
        self.model = csv.read(file)
        # self.model = csv.read("data\\CoP21.csv")

    def test_addActor(self):
        model = EqualGainModel()

        a1 = model.add_actor("TestActor1")
        a2 = model.add_actor("TestActor2")

        self.assertEqual(len(model.actors), 2)
        self.assertNotEqual(a1, a2)

    def test_first_phase(self):
        model = self.model

        model.calc_nbs()  # tested in test_calculations.py#test_calc_nbs

        model.calc_combinations()  # tested below
        model.determine_positions()
        model.determine_groups()

        totalactor_issues = 0

        for k, v in model.groups.items():
            a = len(model.groups[k]["a"])
            b = len(model.groups[k]["b"])
            c = len(model.groups[k]["c"])
            d = len(model.groups[k]["d"])

            self.assertEqual(sum([a, b, c, d]), 10)

            totalactor_issues += a * d
            totalactor_issues += b * c

        self.assertEqual(totalactor_issues, 240)

        e = model.highest_gain()

        self.assertAlmostEqual(
            e.gain, Decimal(1.973684210526315789473684214), delta=1e-8
        )

        # the delta is necessary for the random component by exchanges which have an equal gain.
        # in some cases the gain of the exchanges (two exchanges with each two (unique) actors) are the same.
        # we pick random the next exchange. Theoretically it is even possible to get more exchanges with the same gain
        self.assertAlmostEqual(len(model.exchanges), 239, delta=2)
        realized = list()
        realized.append(e)

        model.remove_invalid_exchanges(e)

        e1 = model.highest_gain()
        model.remove_invalid_exchanges(e1)
        realized.append(e1)

        self.assertAlmostEqual(
            e1.gain, Decimal(1.79263630876534102340553952), delta=1e-8
        )

        while len(model.exchanges) > 0:
            realize = model.highest_gain()
            model.remove_invalid_exchanges(realize)
            realized.append(realize)

            # self.assertEqual(len(realized), 16)

    def test_calc_combinations(self):
        self.model.calc_combinations()
        combinations = len(list(self.model.issue_combinations))

        n = len(self.model.issues)
        k = 2

        self.assertEqual(
            combinations,
            math.factorial(n) / (math.factorial(k) * math.factorial(n - k)),
        )



import os
from unittest import TestCase

from decide.model import equalgain
from decide.model.helpers import csvparser


class TestParser(TestCase):
    def setUp(self):
        self.parser = csvparser.CsvParser(equalgain.EqualGainModel())

    def test_read(self):
        file = os.path.join(
            os.path.dirname(os.path.realpath(__file__)),
            "../../../data/input/sample_data.txt",
        )
        model = self.parser.read(file)

        self.assertEqual(len(model.actors), 10)
        self.assertEqual(len(model.issues), 6)
        self.assertEqual(len(model.actor_issues), 6)
        self.assertEqual(
            len(model.actor_issues[model.issues["tolheffingbinnenstad"]]), 10
        )



from setuptools import setup

from decide import __version__

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="decide-exchange-model",
    version=__version__,
    packages=[
        "decide",
        "decide.qt",
        "decide.qt.inputwindow",
        "decide.model",
        "decide.model.test",
        "decide.model.helpers",
        "decide.model.helpers.test",
        "decide.model.observers",
    ],
    url="https://github.com/foarsitter/decide-exchange-model",
    license="GPL-3.0",
    author="jelmert",
    author_email="info@jelmert.nl",
    description="Model of collective decision making",
    long_description=long_description,
    long_description_content_type="text/markdown",
    entry_points={
        "console_scripts": ["decide-cli=decide.cli:main"],
        "gui_scripts": ["decide-gui=decide.qt.mainwindow:main"],
    },
    #data_files=[('data/input', ['data/input/kopenhagen.csv', 'data/input/CoP21.csv'])],
    python_requires=">=3.6",
    install_requires=["peewee>=3", "matplotlib", "requests", "blinker"],
)



from unittest import TestCase

from decide.model.base import Actor, Issue


class TestIssue(TestCase):
    def test_actor_functions(self):
        actor1 = Actor("Test")
        actor2 = Actor("test")
        actor3 = Actor("Test 123")

        # actor comparison happens on the key attribute. So the following is correct
        self.assertEqual(actor1, actor1)
        self.assertEqual(actor1, actor2)
        self.assertNotEqual(actor1, actor3)

        # naming functions
        self.assertEqual(actor1.name, "Test")
        self.assertEqual(actor1.actor_id, "test")
        self.assertEqual(actor2, "test")
        self.assertEqual(actor3.name, "Test 123")
        self.assertEqual(actor3.actor_id, "test123")
        self.assertEqual(str(actor3), actor3.name)

        self.assertFalse(actor1 is None)

        with self.assertRaises(NotImplementedError):
            self.assertEqual(actor1, Issue("mock"))

        # testing hash index functions
        store = {actor1: actor1, actor2: actor2, actor3: actor3}

        self.assertEqual(store[actor1], actor1)
        self.assertEqual(store[actor2], actor1)
        self.assertNotEqual(store[actor3], actor1)

        self.assertTrue(actor1 < actor3)

        with self.assertRaises(ValueError):
            self.assertTrue(actor1 < Issue("Mock"))
